# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tools for payload serialization and deserialization."""

from enum import Enum
from typing import Any, Optional, Tuple

from packages.valory.contracts.gnosis_safe.contract import SafeOperation


NULL_ADDRESS: str = "0x" + "0" * 40
MAX_UINT256 = 2 ** 256 - 1


class VerificationStatus(Enum):
    """Tx verification status enumeration."""

    VERIFIED = 1
    NOT_VERIFIED = 2
    INVALID_PAYLOAD = 3
    PENDING = 4
    ERROR = 5
    INSUFFICIENT_FUNDS = 6


class PayloadDeserializationError(Exception):
    """Exception for payload deserialization errors."""

    def __init__(self, *args: Any) -> None:
        """Initialize the exception.

        :param args: extra arguments to pass to the constructor of `Exception`.
        """
        msg: str = "Cannot decode provided payload!"
        if not args:
            args = (msg,)

        super().__init__(*args)


def tx_hist_payload_to_hex(
    verification: VerificationStatus, tx_hash: Optional[str] = None
) -> str:
    """Serialise history payload to a hex string."""
    if tx_hash is None:
        tx_hash = ""
    else:
        tx_hash = tx_hash[2:] if tx_hash.startswith("0x") else tx_hash
        if len(tx_hash) != 64:
            raise ValueError("Cannot encode tx_hash of non-32 bytes")
    verification_ = verification.value.to_bytes(32, "big").hex()
    concatenated = verification_ + tx_hash
    return concatenated


def tx_hist_hex_to_payload(payload: str) -> Tuple[VerificationStatus, Optional[str]]:
    """Decode history payload."""
    if len(payload) != 64 and len(payload) != 64 * 2:
        raise PayloadDeserializationError()

    verification_value = int.from_bytes(bytes.fromhex(payload[:64]), "big")

    try:
        verification_status = VerificationStatus(verification_value)
    except ValueError as e:
        raise PayloadDeserializationError(str(e)) from e

    if len(payload) == 64:
        return verification_status, None

    return verification_status, "0x" + payload[64:]


def hash_payload_to_hex(  # pylint: disable=too-many-arguments, too-many-locals
    safe_tx_hash: str,
    ether_value: int,
    safe_tx_gas: int,
    to_address: str,
    data: bytes,
    operation: int = SafeOperation.CALL.value,
    base_gas: int = 0,
    safe_gas_price: int = 0,
    gas_token: str = NULL_ADDRESS,
    refund_receiver: str = NULL_ADDRESS,
) -> str:
    """Serialise to a hex string."""
    if len(safe_tx_hash) != 64:  # should be exactly 32 bytes!
        raise ValueError(
            "cannot encode safe_tx_hash of non-32 bytes"
        )  # pragma: nocover

    if len(to_address) != 42 or len(gas_token) != 42 or len(refund_receiver) != 42:
        raise ValueError("cannot encode address of non 42 length")  # pragma: nocover

    if (
        ether_value > MAX_UINT256
        or safe_tx_gas > MAX_UINT256
        or base_gas > MAX_UINT256
        or safe_gas_price > MAX_UINT256
    ):
        raise ValueError(
            "Value is bigger than the max 256 bit value"
        )  # pragma: nocover

    if operation not in [v.value for v in SafeOperation]:
        raise ValueError("SafeOperation value is not valid")  # pragma: nocover

    ether_value_ = ether_value.to_bytes(32, "big").hex()
    safe_tx_gas_ = safe_tx_gas.to_bytes(32, "big").hex()
    operation_ = operation.to_bytes(1, "big").hex()
    base_gas_ = base_gas.to_bytes(32, "big").hex()
    safe_gas_price_ = safe_gas_price.to_bytes(32, "big").hex()

    concatenated = (
        safe_tx_hash
        + ether_value_
        + safe_tx_gas_
        + to_address
        + operation_
        + base_gas_
        + safe_gas_price_
        + gas_token
        + refund_receiver
        + data.hex()
    )
    return concatenated


def skill_input_hex_to_payload(payload: str) -> dict:
    """Decode payload."""
    if len(payload) < 234:
        raise PayloadDeserializationError()  # pragma: nocover
    tx_params = dict(
        safe_tx_hash=payload[:64],
        ether_value=int.from_bytes(bytes.fromhex(payload[64:128]), "big"),
        safe_tx_gas=int.from_bytes(bytes.fromhex(payload[128:192]), "big"),
        to_address=payload[192:234],
        operation=int.from_bytes(bytes.fromhex(payload[234:236]), "big"),
        base_gas=int.from_bytes(bytes.fromhex(payload[236:300]), "big"),
        safe_gas_price=int.from_bytes(bytes.fromhex(payload[300:364]), "big"),
        gas_token=payload[364:406],
        refund_receiver=payload[406:448],
        data=bytes.fromhex(payload[448:]),
    )
    return tx_params



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the models.py module of the skill."""
import json
import logging
from enum import Enum
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep
from typing import Any, List, Optional, Tuple
from unittest import mock
from unittest.mock import MagicMock

import pytest
from aea.skills.base import SkillContext

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbstractRound,
    BaseSynchronizedData,
)
from packages.valory.skills.abstract_round_abci.models import (
    ApiSpecs,
    BaseParams,
    BenchmarkTool,
    NUMBER_OF_RETRIES,
    Requests,
    SharedState,
)

from tests.test_skills.test_abstract_round_abci.test_base import AbciAppTest


class DummyMessage:
    """Dummy api specs class."""

    body: bytes

    def __init__(self, body: bytes) -> None:
        """Initializes DummyMessage"""
        self.body = body


class TestApiSpecsModel:
    """Test ApiSpecsModel."""

    api_specs: ApiSpecs

    def setup(
        self,
    ) -> None:
        """Setup test."""

        self.api_specs = ApiSpecs(
            name="price_api",
            skill_context=SkillContext(),
            url="http://localhost",
            api_id="api_id",
            method="GET",
            headers=[["Dummy-Header", "dummy_value"]],
            parameters=[["Dummy-Param", "dummy_param"]],
            response_key="value",
            response_type="float",
            retries=NUMBER_OF_RETRIES,
        )

    def test_init(
        self,
    ) -> None:
        """Test initialization."""

        # test ensure method.
        with pytest.raises(ValueError, match="Value for url is required by ApiSpecs"):
            _ = ApiSpecs(
                name="price_api",
                skill_context=SkillContext(),
            )

        assert self.api_specs._retries == NUMBER_OF_RETRIES
        assert self.api_specs._retries_attempted == 0

        assert self.api_specs.url == "http://localhost"
        assert self.api_specs.api_id == "api_id"
        assert self.api_specs.method == "GET"
        assert self.api_specs.headers == [["Dummy-Header", "dummy_value"]]
        assert self.api_specs.parameters == [["Dummy-Param", "dummy_param"]]
        assert self.api_specs.response_key == "value"
        assert self.api_specs.response_type == "float"

    def test_retries(
        self,
    ) -> None:
        """Tests for retries."""

        self.api_specs.increment_retries()
        assert self.api_specs._retries_attempted == 1
        assert not self.api_specs.is_retries_exceeded()

        for _ in range(NUMBER_OF_RETRIES):
            self.api_specs.increment_retries()
        assert self.api_specs.is_retries_exceeded()
        self.api_specs.reset_retries()
        assert self.api_specs._retries_attempted == 0

    def test_get_spec(
        self,
    ) -> None:
        """Test get_spec method."""

        actual_specs = {
            "url": "http://localhost",
            "method": "GET",
            "headers": [["Dummy-Header", "dummy_value"]],
            "parameters": [["Dummy-Param", "dummy_param"]],
        }

        specs = self.api_specs.get_spec()
        assert all([key in specs for key in actual_specs.keys()])
        assert all([specs[key] == actual_specs[key] for key in actual_specs])

    def test_process_response_with_depth_0(
        self,
    ) -> None:
        """Test process_response method."""

        value = self.api_specs.process_response(DummyMessage(b""))  # type: ignore
        assert value is None

        value = self.api_specs.process_response(
            DummyMessage(b'{"value": "10.232"}')  # type: ignore
        )
        assert isinstance(value, float)

    def test_process_response_with_depth_1(
        self,
    ) -> None:
        """Test process_response method."""

        api_specs = ApiSpecs(
            name="price_api",
            skill_context=SkillContext(),
            url="http://localhost",
            api_id="api_id",
            method="GET",
            headers="Dummy-Header:dummy_value",
            parameters="Dummy-Param:dummy_param",
            response_key="value_0:value_1",
            response_type="float",
            retries=NUMBER_OF_RETRIES,
        )

        value = api_specs.process_response(
            DummyMessage(b'{"value_0": {"value_1": "10.232"}}')  # type: ignore
        )
        assert isinstance(value, float)

    def test_process_response_with_key_none(
        self,
    ) -> None:
        """Test process_response method."""

        api_specs = ApiSpecs(
            name="price_api",
            skill_context=SkillContext(),
            url="http://localhost",
            api_id="api_id",
            method="GET",
            headers="Dummy-Header:dummy_value",
            parameters="Dummy-Param:dummy_param",
            response_key=None,
            response_type=None,
            retries=NUMBER_OF_RETRIES,
        )

        value = api_specs.process_response(
            DummyMessage(b'{"value": "10.232"}')  # type: ignore
        )
        assert isinstance(value, dict)


class ConcreteRound(AbstractRound):
    """A ConcreteRoundA for testing purposes."""

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Handle the end of the block."""


class TestSharedState:
    """Test SharedState(Model) class."""

    @mock.patch.object(SharedState, "_process_abci_app_cls")
    def test_initialization(self, *_: Any) -> None:
        """Test the initialization of the shared state."""
        SharedState(abci_app_cls=MagicMock(), name="", skill_context=MagicMock())

    @mock.patch.object(SharedState, "_process_abci_app_cls")
    def test_setup(self, *_: Any) -> None:
        """Test setup method."""
        shared_state = SharedState(
            abci_app_cls=MagicMock, name="", skill_context=MagicMock()
        )
        shared_state.context.params.setup_params = {"test": []}
        shared_state.context.params.consensus_params = MagicMock()
        shared_state.setup()

    @mock.patch.object(SharedState, "_process_abci_app_cls")
    def test_synchronized_data_negative_not_available(self, *_: Any) -> None:
        """Test 'synchronized_data' property getter, negative case (not available)."""
        shared_state = SharedState(
            abci_app_cls=AbciAppTest, name="", skill_context=MagicMock()
        )
        with mock.patch.object(shared_state.context, "params"):
            with pytest.raises(ValueError, match="round sequence not available"):
                shared_state.synchronized_data

    @mock.patch.object(SharedState, "_process_abci_app_cls")
    def test_synchronized_data_positive(self, *_: Any) -> None:
        """Test 'synchronized_data' property getter, negative case (not available)."""
        shared_state = SharedState(
            abci_app_cls=AbciAppTest, name="", skill_context=MagicMock()
        )
        shared_state.context.params.setup_params = {"test": []}
        shared_state.context.params.consensus_params = MagicMock()
        shared_state.setup()
        shared_state.round_sequence.abci_app._round_results = [MagicMock()]
        shared_state.synchronized_data

    def test_synchronized_data_db(self, *_: Any) -> None:
        """Test 'synchronized_data' AbciAppDB."""
        shared_state = SharedState(
            abci_app_cls=AbciAppTest, name="", skill_context=MagicMock()
        )
        with mock.patch.object(shared_state.context, "params") as mock_params:
            mock_params.setup_params = {
                "safe_contract_address": ["0xsafe"],
                "oracle_contract_address": ["0xoracle"],
            }
            shared_state.setup()
            assert (
                shared_state.synchronized_data.db.get_strict("safe_contract_address")
                == "0xsafe"
            )
            assert (
                shared_state.synchronized_data.db.get_strict("oracle_contract_address")
                == "0xoracle"
            )

    def test_process_abci_app_cls_negative_not_a_class(self) -> None:
        """Test '_process_abci_app_cls', negative case (not a class)."""
        mock_obj = MagicMock()
        with pytest.raises(ValueError, match=f"The object {mock_obj} is not a class"):
            SharedState._process_abci_app_cls(mock_obj)

    def test_process_abci_app_cls_negative_not_subclass_of_abstract_round(
        self,
    ) -> None:
        """Test '_process_abci_app_cls', negative case (not subclass of AbstractRound)."""
        with pytest.raises(
            ValueError,
            match=f"The class {MagicMock} is not an instance of {AbciApp.__module__}.{AbciApp.__name__}",
        ):
            SharedState._process_abci_app_cls(MagicMock)

    def test_process_abci_app_cls_positive(self) -> None:
        """Test '_process_abci_app_cls', positive case."""

        SharedState._process_abci_app_cls(AbciAppTest)


class TestBenchmarkTool:
    """Test BenchmarkTool"""

    @staticmethod
    def _check_behaviour_data(data: List, agent_name: str) -> None:
        """Check behaviour data."""
        assert len(data) == 1

        (behaviour_data,) = data
        assert behaviour_data["behaviour"] == agent_name
        assert all(
            [key in behaviour_data["data"] for key in ("local", "consensus", "total")]
        )

    def test_end_2_end(self) -> None:
        """Test end 2 end of the tool."""

        agent_name = "agent"
        skill_context = MagicMock(
            agent_address=agent_name, logger=MagicMock(info=logging.info)
        )

        with TemporaryDirectory() as temp_dir:
            benchmark = BenchmarkTool(
                name=agent_name, skill_context=skill_context, log_dir=temp_dir
            )

            with benchmark.measure(agent_name).local():
                sleep(1.0)

            with benchmark.measure(agent_name).consensus():
                sleep(1.0)

            self._check_behaviour_data(benchmark.data, agent_name)

            benchmark.save()

            benchmark_dir = Path(temp_dir, agent_name)
            benchmark_file = benchmark_dir / "0.json"
            assert (benchmark_file).is_file()

            behaviour_data = json.loads(benchmark_file.read_text())
            self._check_behaviour_data(behaviour_data, agent_name)


def test_requests_model_initialization() -> None:
    """Test initialization of the 'Requests(Model)' class."""
    Requests(name="", skill_context=MagicMock())


def test_base_params_model_initialization() -> None:
    """Test initialization of the 'BaseParams(Model)' class."""
    BaseParams(
        name="",
        skill_context=MagicMock(),
        setup={},
        consensus=dict(max_participants=1),
        tendermint_url="",
        max_healthcheck=1,
        round_timeout_seconds=1,
        sleep_time=1,
        retry_timeout=1,
        retry_attempts=1,
        observation_interval=1,
        drand_public_key="",
        tendermint_com_url="",
        reset_tendermint_after=1,
        service_id="abstract_round_abci",
        keeper_timeout=1.0,
        tendermint_check_sleep_delay=3,
        tendermint_max_retries=5,
        cleanup_history_depth=0,
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/reset_pause_abci skill's behaviours."""
from pathlib import Path
from typing import Callable, Generator, Optional
from unittest import mock
from unittest.mock import MagicMock

import pytest

from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.base import (
    BaseSynchronizedData as ResetSynchronizedSata,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    make_degenerate_behaviour,
)
from packages.valory.skills.reset_pause_abci.behaviours import ResetAndPauseBehaviour
from packages.valory.skills.reset_pause_abci.rounds import Event as ResetEvent
from packages.valory.skills.reset_pause_abci.rounds import FinishedResetAndPauseRound

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase


class ResetPauseAbciFSMBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing PauseReset FSMBehaviour."""

    path_to_skill = Path(ROOT_DIR, "packages", "valory", "skills", "reset_pause_abci")


def dummy_reset_tendermint_with_wait_wrapper(
    reset_successfully: Optional[bool],
) -> Callable[[], Generator[None, None, Optional[bool]]]:
    """Wrapper for a Dummy `reset_tendermint_with_wait` method."""

    def dummy_reset_tendermint_with_wait() -> Generator[None, None, Optional[bool]]:
        """Dummy `reset_tendermint_with_wait` method."""
        yield
        return reset_successfully

    return dummy_reset_tendermint_with_wait


class TestResetAndPauseBehaviour(ResetPauseAbciFSMBehaviourBaseCase):
    """Test ResetBehaviour."""

    behaviour_class = ResetAndPauseBehaviour
    next_behaviour_class = make_degenerate_behaviour(
        FinishedResetAndPauseRound.round_id
    )

    @pytest.mark.parametrize("tendermint_reset_status", (None, True, False))
    def test_reset_behaviour(
        self,
        tendermint_reset_status: Optional[bool],
    ) -> None:
        """Test reset behaviour."""

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.behaviour_id,
            synchronized_data=ResetSynchronizedSata(
                AbciAppDB(
                    initial_data=dict(
                        most_voted_estimate=[0.1],
                        tx_hashes_history=[["68656c6c6f776f726c64"]],
                    ),
                )
            ),
        )

        assert self.behaviour.current_behaviour is not None
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.behaviour_id
        )

        with mock.patch.object(
            self.behaviour.current_behaviour,
            "send_a2a_transaction",
            side_effect=self.behaviour.current_behaviour.send_a2a_transaction,
        ), mock.patch.object(
            self.behaviour.current_behaviour,
            "reset_tendermint_with_wait",
            side_effect=dummy_reset_tendermint_with_wait_wrapper(
                tendermint_reset_status
            ),
        ) as mock_reset_tendermint_with_wait, mock.patch.object(
            self.behaviour.current_behaviour,
            "wait_from_last_timestamp",
            side_effect=lambda _: (yield),
        ):
            if tendermint_reset_status is not None:
                # Increase the period_count to force the call to reset_tendermint_with_wait()
                self.behaviour.current_behaviour.synchronized_data.create()

            self.behaviour.act_wrapper()
            self.behaviour.act_wrapper()

            # now if the first reset attempt has been simulated to fail, let's simulate the second attempt to succeed.
            if tendermint_reset_status is not None and not tendermint_reset_status:
                mock_reset_tendermint_with_wait.side_effect = (
                    dummy_reset_tendermint_with_wait_wrapper(True)
                )
                self.behaviour.act_wrapper()
                # make sure that the behaviour does not send any txs to other agents when Tendermint reset fails
                assert isinstance(
                    self.behaviour.current_behaviour.send_a2a_transaction, MagicMock
                )
                self.behaviour.current_behaviour.send_a2a_transaction.assert_not_called()
                self.behaviour.act_wrapper()

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(ResetEvent.DONE)
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.next_behaviour_class.behaviour_id
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abci' skill."""

from abc import ABC
from decimal import Decimal
from typing import Dict, Generator, Optional, Sequence, Set, Type, cast

from aea.exceptions import enforce

from packages.valory.contracts.gnosis_safe.contract import GnosisSafeContract
from packages.valory.contracts.offchain_aggregator.contract import (
    OffchainAggregatorContract,
)
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.price_estimation_abci.models import Params
from packages.valory.skills.price_estimation_abci.payloads import (
    EstimatePayload,
    ObservationPayload,
    TransactionHashPayload,
)
from packages.valory.skills.price_estimation_abci.rounds import (
    CollectObservationRound,
    EstimateConsensusRound,
    PriceAggregationAbciApp,
    SynchronizedData,
    TxHashRound,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    hash_payload_to_hex,
)


# This safeTxGas value is calculated from experimental values plus
# a 10% buffer and rounded up. The Gnosis safe default value is 0 (max gas)
# https://help.gnosis-safe.io/en/articles/4738445-advanced-transaction-parameters
# More on gas estimation: https://help.gnosis-safe.io/en/articles/4933491-gas-estimation
SAFE_TX_GAS = 120000
ETHER_VALUE = 0

NO_OBSERVATION = 0.0


def to_int(most_voted_estimate: float, decimals: int) -> int:
    """Convert to int."""
    most_voted_estimate_ = str(most_voted_estimate)
    decimal_places = most_voted_estimate_[::-1].find(".")
    if decimal_places > decimals:
        most_voted_estimate_ = most_voted_estimate_[: -(decimal_places - decimals)]
    most_voted_estimate_decimal = Decimal(most_voted_estimate_)
    int_value = int(most_voted_estimate_decimal * (10 ** decimals))
    return int_value


class PriceEstimationBaseBehaviour(BaseBehaviour, ABC):
    """Base behaviour for the common apps' skill."""

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, super().synchronized_data)

    @property
    def params(self) -> Params:
        """Return the params."""
        return cast(Params, super().params)


class ObserveBehaviour(PriceEstimationBaseBehaviour):
    """Observe price estimate."""

    behaviour_id = "collect_observation"
    matching_round = CollectObservationRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Ask the configured API the price of a currency.
        - If the request fails, retry until max retries are exceeded.
        - Send an observation transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        if self.context.price_api.is_retries_exceeded():
            # now we need to wait and see if the other agents progress the round, otherwise we should restart?
            with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
                yield from self.wait_until_round_end()
            self.set_done()
            return

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            api_specs = self.context.price_api.get_spec()
            response = yield from self.get_http_response(
                method=api_specs["method"],
                url=api_specs["url"],
                headers=api_specs["headers"],
                parameters=api_specs["parameters"],
            )
            observation = self.context.price_api.process_response(response)

        if observation:
            self.context.logger.info(
                f"Got observation of {self.context.price_api.currency_id} price in "
                + f"{self.context.price_api.convert_id} from {self.context.price_api.api_id}: "
                + f"{observation}"
            )
            payload = ObservationPayload(self.context.agent_address, observation)
            with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
                yield from self.send_a2a_transaction(payload)
                yield from self.wait_until_round_end()
            self.set_done()
        else:
            self.context.logger.info(
                f"Could not get price from {self.context.price_api.api_id}"
            )
            yield from self.sleep(self.params.sleep_time)
            self.context.price_api.increment_retries()

    def clean_up(self) -> None:
        """
        Clean up the resources due to a 'stop' event.

        It can be optionally implemented by the concrete classes.
        """
        self.context.price_api.reset_retries()


class EstimateBehaviour(PriceEstimationBaseBehaviour):
    """Estimate price."""

    behaviour_id = "estimate"
    matching_round = EstimateConsensusRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Run the script to compute the estimate starting from the shared observations.
        - Build an estimate transaction and send the transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():

            self.synchronized_data.set_aggregator_method(
                self.params.observation_aggregator_function
            )
            self.context.logger.info(
                "Got estimate of %s price in %s: %s, Using aggregator method: %s",
                self.context.price_api.currency_id,
                self.context.price_api.convert_id,
                self.synchronized_data.estimate,
                self.params.observation_aggregator_function,
            )
            payload = EstimatePayload(
                self.context.agent_address, self.synchronized_data.estimate
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


def pack_for_server(  # pylint: disable-msg=too-many-arguments
    participants: Sequence[str],
    decimals: int,
    period_count: int,
    estimate: float,
    observations: Dict[str, float],
    data_source: str,
    unit: str,
    **_: Dict[str, str],
) -> bytes:
    """Package server data for signing"""
    enforce(len(str(estimate)) <= 32, "'estimate' too large")
    enforce(len(data_source) <= 32, "'data_source' too large")
    enforce(len(unit) <= 32, "'unit' too large")
    enforce(
        all(len(str(value)) <= 32 for value in observations.values()),
        "'observation' values too large",
    )
    observed = (
        to_int(observations.get(p, NO_OBSERVATION), decimals) for p in participants
    )
    return b"".join(
        [
            period_count.to_bytes(32, "big"),
            to_int(estimate, decimals).to_bytes(32, "big"),
            *map(lambda n: n.to_bytes(32, "big"), observed),
            str(data_source).zfill(32).encode("utf-8"),
            str(unit).zfill(32).encode("utf-8"),
        ]
    )


class TransactionHashBehaviour(PriceEstimationBaseBehaviour):
    """Share the transaction hash for the signature round."""

    behaviour_id = "tx_hash"
    matching_round = TxHashRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Request the transaction hash for the safe transaction. This is the
          hash that needs to be signed by a threshold of agents.
        - Send the transaction hash as a transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            payload_string = yield from self._get_safe_tx_hash()
            payload = TransactionHashPayload(self.context.agent_address, payload_string)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            if self.params.is_broadcasting_to_server:
                yield from self.send_to_server()
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def send_to_server(self) -> Generator:  # pylint: disable-msg=too-many-locals
        """
        Send data to server.

        We send current period data of the agents and the previous
        cycle's on-chain settlement tx hash. The current cycle's tx hash
        is not available at this stage yet, and the first iteration will
        contain no tx hash since there has not been on-chain transaction
        settlement yet.

        :yield: the http response
        """

        period_count = self.synchronized_data.period_count

        self.context.logger.info("Attempting broadcast")

        prev_tx_hash = ""
        if period_count != 0:
            # grab tx_hash from previous cycle
            prev_period_count = period_count - 1
            previous_data = self.synchronized_data.db.get_latest_from_reset_index(
                prev_period_count
            )

            prev_tx_hash = previous_data.get("final_tx_hash", "")

        # select relevant data
        agents = self.synchronized_data.db.get_strict("participants")
        payloads = self.synchronized_data.db.get_strict("participant_to_observations")
        estimate = self.synchronized_data.db.get_strict("most_voted_estimate")

        observations = {
            agent: getattr(payloads.get(agent), "observation", NO_OBSERVATION)
            for agent in agents
        }

        price_api = self.context.price_api

        # adding timestamp on server side when received
        # period and agent_address are used as `primary key`
        data_for_server = {
            "period_count": period_count,
            "agent_address": self.context.agent_address,
            "estimate": estimate,
            "prev_tx_hash": prev_tx_hash,
            "observations": observations,
            "data_source": price_api.api_id,
            "unit": f"{price_api.currency_id}:{price_api.convert_id}",
        }

        # pack data
        participants = self.synchronized_data.sorted_participants
        decimals = self.params.oracle_params["decimals"]
        package = pack_for_server(participants, decimals, **data_for_server)
        data_for_server["package"] = package.hex()

        # get signature
        signature = yield from self.get_signature(package, is_deprecated_mode=True)
        data_for_server["signature"] = signature
        self.context.logger.info(f"Signature: {signature}")

        message = str(data_for_server).encode("utf-8")
        server_api_specs = self.context.server_api.get_spec()
        raw_response = yield from self.get_http_response(
            method=server_api_specs["method"],
            url=server_api_specs["url"],
            content=message,
        )

        response = self.context.server_api.process_response(raw_response)
        self.context.logger.info(f"Broadcast response: {response}")

    def _get_safe_tx_hash(self) -> Generator[None, None, Optional[str]]:
        """Get the transaction hash of the Safe tx."""
        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.oracle_contract_address,
            contract_id=str(OffchainAggregatorContract.contract_id),
            contract_callable="get_latest_transmission_details",
        )
        if (
            contract_api_msg.performative
            != ContractApiMessage.Performative.RAW_TRANSACTION
        ):  # pragma: nocover
            self.context.logger.warning("get_latest_transmission_details unsuccessful!")
            return None
        epoch_ = cast(int, contract_api_msg.raw_transaction.body["epoch_"]) + 1
        round_ = cast(int, contract_api_msg.raw_transaction.body["round_"])
        decimals = self.params.oracle_params["decimals"]
        amount = self.synchronized_data.most_voted_estimate
        amount_ = to_int(amount, decimals)
        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.oracle_contract_address,
            contract_id=str(OffchainAggregatorContract.contract_id),
            contract_callable="get_transmit_data",
            epoch_=epoch_,
            round_=round_,
            amount_=amount_,
        )
        if (
            contract_api_msg.performative
            != ContractApiMessage.Performative.RAW_TRANSACTION
        ):  # pragma: nocover
            self.context.logger.warning("get_transmit_data unsuccessful!")
            return None
        data = cast(bytes, contract_api_msg.raw_transaction.body["data"])
        to_address = self.synchronized_data.oracle_contract_address
        ether_value = ETHER_VALUE
        safe_tx_gas = SAFE_TX_GAS
        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_raw_safe_transaction_hash",
            to_address=to_address,
            value=ether_value,
            data=data,
            safe_tx_gas=safe_tx_gas,
        )
        if (
            contract_api_msg.performative
            != ContractApiMessage.Performative.RAW_TRANSACTION
        ):  # pragma: nocover
            self.context.logger.warning("get_raw_safe_transaction_hash unsuccessful!")
            return None
        safe_tx_hash = cast(str, contract_api_msg.raw_transaction.body["tx_hash"])
        safe_tx_hash = safe_tx_hash[2:]
        self.context.logger.info(f"Hash of the Safe transaction: {safe_tx_hash}")
        # temp hack:
        payload_string = hash_payload_to_hex(
            safe_tx_hash, ether_value, safe_tx_gas, to_address, data
        )
        return payload_string


class ObserverRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the observer behaviour."""

    initial_behaviour_cls = ObserveBehaviour
    abci_app_cls = PriceAggregationAbciApp  # type: ignore
    behaviours: Set[Type[BaseBehaviour]] = {  # type: ignore
        ObserveBehaviour,  # type: ignore
        EstimateBehaviour,  # type: ignore
        TransactionHashBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the class to connect to an Gnosis Safe contract."""
# heavily borrows from https://github.com/gnosis/gnosis-py/blob/51b41f5a8577a96e296b9b7e037491632cda9d8c/gnosis/safe/multi_send.py
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, cast

from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi
from hexbytes import HexBytes
from web3 import Web3


PUBLIC_ID = PublicId.from_str("valory/multisend:0.1.0")

_logger = logging.getLogger(
    f"aea.packages.{PUBLIC_ID.author}.contracts.{PUBLIC_ID.name}.contract"
)


class MultiSendOperation(Enum):
    """Operation types."""

    CALL = 0
    DELEGATE_CALL = 1


def encode_data(tx: Dict) -> bytes:
    """Encodes multisend transaction."""
    operation = HexBytes(
        "{:0>2x}".format(cast(MultiSendOperation, tx.get("operation")).value)
    )  # Operation 1 byte
    to = HexBytes(
        "{:0>40x}".format(int(cast(str, tx.get("to")), 16))
    )  # Address 20 bytes
    value = HexBytes("{:0>64x}".format(cast(int, tx.get("value"))))  # Value 32 bytes
    data = cast(bytes, tx.get("data", b""))
    data_ = HexBytes(data)
    data_length = HexBytes("{:0>64x}".format(len(data_)))  # Data length 32 bytes
    return operation + to + value + data_length + data_


def decode_data(encoded_tx: bytes) -> Tuple[Dict, int]:
    """Decodes multisend transaction."""
    encoded_tx = HexBytes(encoded_tx)
    operation = MultiSendOperation(encoded_tx[0])
    to = Web3.toChecksumAddress(encoded_tx[1 : 1 + 20])
    value = int.from_bytes(encoded_tx[21 : 21 + 32], byteorder="big")
    data_length = int.from_bytes(encoded_tx[21 + 32 : 21 + 32 * 2], byteorder="big")
    data = encoded_tx[21 + 32 * 2 : 21 + 32 * 2 + data_length]
    len_data = len(data)
    if len_data != data_length:  # pragma: nocover
        raise ValueError(
            f"Data length {data_length} is different from len(data) {len_data}"
        )
    total_length = 21 + 32 * 2 + data_length
    return (
        {"operation": operation, "to": to, "value": value, "data": data},
        total_length,
    )


def to_bytes(multi_send_txs: List[Dict]) -> bytes:
    """Multi send tx list to bytes."""
    return b"".join([encode_data(tx) for tx in multi_send_txs])


def from_bytes(encoded_multisend_txs: bytes) -> List[Dict]:
    """Encoded multi send tx to list."""
    encoded_multisend_txs = HexBytes(encoded_multisend_txs)
    next_data_position = 0
    txs = []
    while len(encoded_multisend_txs[next_data_position:]) > 0:
        tx, total_length = decode_data(encoded_multisend_txs[next_data_position:])
        next_data_position += total_length
        txs.append(tx)
    return txs


class MultiSendContract(Contract):
    """The MultiSend contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get the Safe transaction."""
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[bytes]:
        """Get raw message."""
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get state."""
        raise NotImplementedError

    @classmethod
    def get_deploy_transaction(
        cls, ledger_api: LedgerApi, deployer_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get deploy transaction."""
        raise NotImplementedError

    @classmethod
    def get_tx_data(
        cls, ledger_api: LedgerApi, contract_address: str, multi_send_txs: List[Dict]
    ) -> Optional[JSONLike]:
        """
        Get a multisend transaction data from list.

        :param ledger_api: ledger API object.
        :param contract_address: the contract address.
        :param multi_send_txs: the multisend transaction list.
        :return: an optional JSON-like object.
        """
        multisend_contract = cls.get_instance(ledger_api, contract_address)
        encoded_multisend_data = to_bytes(multi_send_txs)
        return {
            "data": multisend_contract.functions.multiSend(
                encoded_multisend_data
            ).buildTransaction({"gas": 1, "gasPrice": 1})["data"]
        }

    @classmethod
    def get_tx_list(
        cls, ledger_api: LedgerApi, contract_address: str, multi_send_data: str
    ) -> Optional[JSONLike]:
        """
        Get a multisend transaction list from encoded data.

        :param ledger_api: ledger API object.
        :param contract_address: the contract address.
        :param multi_send_data: the multisend transaction data.
        :return: an optional JSON-like object.
        """
        multisend_contract = cls.get_instance(ledger_api, contract_address)
        _, encoded_multisend_data = multisend_contract.decode_function_input(
            multi_send_data
        )
        return {"tx_list": from_bytes(encoded_multisend_data["transactions"])}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the utils.py module of the skill."""
import sys
from unittest import mock

import atheris  # type: ignore
import pytest

from packages.valory.skills.abstract_round_abci.utils import MAX_UINT64, VerifyDrand


DRAND_PUBLIC_KEY: str = "868f005eb8e6e4ca0a47c8a77ceaa5309a47978a7c71bc5cce96366b5d7a569937c529eeda66c7293784a9402801af31"

DRAND_VALUE = {
    "round": 1416669,
    "randomness": "f6be4bf1fa229f22340c1a5b258f809ac4af558200775a67dacb05f0cb258a11",
    "signature": (
        "b44d00516f46da3a503f9559a634869b6dc2e5d839e46ec61a090e3032172954929a5"
        "d9bd7197d7739fe55db770543c71182562bd0ad20922eb4fe6b8a1062ed21df3b68de"
        "44694eb4f20b35262fa9d63aa80ad3f6172dd4d33a663f21179604"
    ),
    "previous_signature": (
        "903c60a4b937a804001032499a855025573040cb86017c38e2b1c3725286756ce8f33"
        "61188789c17336beaf3f9dbf84b0ad3c86add187987a9a0685bc5a303e37b008fba8c"
        "44f02a416480dd117a3ff8b8075b1b7362c58af195573623187463"
    ),
}


class TestVerifyDrand:
    """Test DrandVerify."""

    drand_check: VerifyDrand

    def setup(
        self,
    ) -> None:
        """Setup test."""
        self.drand_check = VerifyDrand()

    def test_verify(
        self,
    ) -> None:
        """Test verify method."""

        result, error = self.drand_check.verify(DRAND_VALUE, DRAND_PUBLIC_KEY)
        assert result
        assert error is None

    def test_verify_fails(
        self,
    ) -> None:
        """Test verify method."""

        drand_value = DRAND_VALUE.copy()
        del drand_value["randomness"]
        result, error = self.drand_check.verify(drand_value, DRAND_PUBLIC_KEY)
        assert not result
        assert error == "DRAND dict is missing value for 'randomness'"

        drand_value = DRAND_VALUE.copy()
        drand_value["randomness"] = "".join(
            list(drand_value["randomness"])[:-1] + ["0"]  # type: ignore
        )
        result, error = self.drand_check.verify(drand_value, DRAND_PUBLIC_KEY)
        assert not result
        assert error == "Failed randomness hash check."

        drand_value = DRAND_VALUE.copy()
        with mock.patch.object(
            self.drand_check, "_verify_signature", return_value=False
        ):
            result, error = self.drand_check.verify(drand_value, DRAND_PUBLIC_KEY)

        assert not result
        assert error == "Failed bls.Verify check."

    @pytest.mark.parametrize("value", (-1, MAX_UINT64 + 1))
    def test_negative_and_overflow(self, value: int) -> None:
        """Test verify method."""
        with pytest.raises(ValueError):
            self.drand_check._int_to_bytes_big(value)


@pytest.mark.skip
def test_fuzz_verify_drand() -> None:
    """Fuzz test for VerifyDrand. Run directly as a function, not through pytest"""

    @atheris.instrument_func
    def test_verify_int_to_bytes_big(input_bytes: bytes) -> None:
        """Test VerifyDrand."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        VerifyDrand._int_to_bytes_big(
            fdp.ConsumeInt(16)
        )  # pylint: disable=protected-access

    @atheris.instrument_func
    def test_verify_randomness_hash(input_bytes: bytes) -> None:
        """Test VerifyDrand."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        VerifyDrand._verify_randomness_hash(
            fdp.ConsumeBytes(16), fdp.ConsumeBytes(16)
        )  # pylint: disable=protected-access

    atheris.instrument_all()
    atheris.Setup(sys.argv, test_verify_int_to_bytes_big)
    atheris.Fuzz()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for the valory/oracle_abci skill."""
import pytest
from aea.configurations.data_types import PublicId

from tests.fixture_helpers import UseGnosisSafeHardHatNet
from tests.test_agents.base import (
    BaseTestEnd2EndAgentCatchup,
    BaseTestEnd2EndNormalExecution,
)


# round check log messages of the happy path
EXPECTED_ROUND_LOG_COUNT = {
    "registration_startup": 1,
    "randomness_safe": 1,
    "select_keeper_safe": 1,
    "deploy_safe": 1,
    "validate_safe": 1,
    "randomness_oracle": 1,
    "select_keeper_oracle": 1,
    "deploy_oracle": 1,
    "validate_oracle": 1,
    "estimate_consensus": 2,
    "tx_hash": 2,
    "randomness_transaction_submission": 2,
    "select_keeper_transaction_submission_a": 2,
    "collect_signature": 2,
    "finalization": 2,
    "validate_transaction": 2,
    "reset_and_pause": 2,
    "collect_observation": 3,
}

# strict check log messages of the happy path
STRICT_CHECK_STRINGS = (
    "Finalized with transaction hash",
    "Signature:",
    "Got estimate of BTC price in USD:",
    "Got observation of BTC price in USD",
    "Period end",
)


@pytest.mark.parametrize("nb_nodes", (1,))
class TestABCIPriceEstimationSingleAgent(
    BaseTestEnd2EndNormalExecution,
    UseGnosisSafeHardHatNet,
):
    """Test the ABCI oracle skill with only one agent."""

    agent_package = "valory/oracle:0.1.0"
    skill_package = "valory/oracle_abci:0.1.0"
    wait_to_finish = 180
    strict_check_strings = STRICT_CHECK_STRINGS
    round_check_strings_to_n_periods = EXPECTED_ROUND_LOG_COUNT


@pytest.mark.parametrize("nb_nodes", (2,))
class TestABCIPriceEstimationTwoAgents(
    BaseTestEnd2EndNormalExecution,
    UseGnosisSafeHardHatNet,
):
    """Test the ABCI oracle skill with two agents."""

    agent_package = "valory/oracle:0.1.0"
    skill_package = "valory/oracle_abci:0.1.0"
    wait_to_finish = 180
    strict_check_strings = STRICT_CHECK_STRINGS
    round_check_strings_to_n_periods = EXPECTED_ROUND_LOG_COUNT


@pytest.mark.parametrize("nb_nodes", (4,))
class TestABCIPriceEstimationFourAgents(
    BaseTestEnd2EndNormalExecution,
    UseGnosisSafeHardHatNet,
):
    """Test the ABCI oracle skill with four agents."""

    agent_package = "valory/oracle:0.1.0"
    skill_package = "valory/oracle_abci:0.1.0"
    wait_to_finish = 180
    strict_check_strings = STRICT_CHECK_STRINGS
    round_check_strings_to_n_periods = EXPECTED_ROUND_LOG_COUNT


@pytest.mark.parametrize("nb_nodes", (4,))
class TestAgentCatchup(BaseTestEnd2EndAgentCatchup, UseGnosisSafeHardHatNet):
    """Test that an agent that is launched later can synchronize with the rest of the network"""

    agent_package = "valory/oracle:0.1.0"
    skill_package = "valory/oracle_abci:0.1.0"
    wait_to_finish = 200
    restart_after = 45
    round_check_strings_to_n_periods = EXPECTED_ROUND_LOG_COUNT
    stop_string = "'registration_startup' round is done with event: Event.DONE"


@pytest.mark.skip
class TestTendermintReset(TestABCIPriceEstimationFourAgents):
    """Test the ABCI oracle skill with four agents when resetting Tendermint."""

    skill_package = "valory/oracle_abci:0.1.0"
    wait_to_finish = 360
    # run for 4 periods instead of 2
    round_check_strings_to_n_periods = {
        round: 4
        for round in EXPECTED_ROUND_LOG_COUNT.keys()
        if round
        in (
            "estimate_consensus"
            "tx_hash"
            "randomness_transaction_submission"
            "select_keeper_transaction_submission_a"
            "collect_signature"
            "finalization"
            "validate_transaction"
            "reset_and_pause"
            "collect_observation"
        )
    }
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    # reset every two rounds
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": 2,
        },
    ]


@pytest.mark.skip
class TestTendermintResetInterrupt(TestAgentCatchup):
    """Test the ABCI oracle skill with four agents when an agent gets temporarily interrupted on Tendermint reset."""

    skill_package = "valory/oracle_abci:0.1.0"
    cli_log_options = ["-v", "INFO"]
    wait_before_stop = 100
    wait_to_finish = 300
    restart_after = 1
    __n_resets_to_perform = 3
    __reset_tendermint_every = 2

    # stop for restart_after seconds when resetting Tendermint for the first time (using -1 because count starts from 0)
    stop_string = f"Entered in the 'reset_and_pause' round for period {__reset_tendermint_every - 1}"
    # check if we manage to reset with Tendermint `__n_resets_to_perform` times with the rest of the agents
    exclude_from_checks = [3]
    round_check_strings_to_n_periods = {
        "reset_and_pause": __n_resets_to_perform * __reset_tendermint_every
    }
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    # reset every `__reset_tendermint_every` rounds
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": __reset_tendermint_every,
        },
    ]


@pytest.mark.skip
class TestTendermintResetInterruptNoRejoin(TestTendermintResetInterrupt):
    """
    Test a Tendermint reset case for the ABCI oracle skill.

    Test the ABCI oracle skill with four agents when an agent gets temporarily interrupted
    on Tendermint reset and never rejoins.
    """

    wait_to_finish = 300
    # set the restart to a value so that the agent never rejoins, in order to test the impact to the rest of the agents
    restart_after = wait_to_finish



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Script to create environment for benchmarking n agents."""

from pathlib import Path
from typing import Any, Dict, List, cast

import yaml

from autonomy.constants import (
    HARDHAT_IMAGE_NAME,
    OPEN_AEA_IMAGE_NAME,
    TENDERMINT_IMAGE_NAME,
    TENDERMINT_IMAGE_VERSION,
)
from autonomy.deploy.base import BaseDeploymentGenerator, ServiceSpecification
from autonomy.deploy.constants import (
    DEFAULT_ENCODING,
    KUBERNETES_AGENT_KEY_NAME,
    TENDERMINT_CONFIGURATION_OVERRIDES,
)
from autonomy.deploy.generators.kubernetes.templates import (
    AGENT_NODE_TEMPLATE,
    AGENT_SECRET_TEMPLATE,
    CLUSTER_CONFIGURATION_TEMPLATE,
    HARDHAT_TEMPLATE,
)


class KubernetesGenerator(BaseDeploymentGenerator):
    """Kubernetes Deployment Generator."""

    image_name: str
    output_name: str = "build.yaml"
    deployment_type: str = "kubernetes"

    def __init__(self, service_spec: ServiceSpecification, build_dir: Path) -> None:
        """Initialise the deployment generator."""
        super().__init__(service_spec, build_dir)
        self.resources: List[str] = []

    def build_agent_deployment(
        self,
        image_name: str,
        agent_ix: int,
        number_of_agents: int,
        agent_vars: Dict[str, Any],
        image_versions: Dict[str, str],
    ) -> str:
        """Build agent deployment."""

        host_names = ", ".join(
            [f'"--hostname=abci{i}"' for i in range(number_of_agents)]
        )

        agent_deployment = AGENT_NODE_TEMPLATE.format(
            valory_app=image_name,
            validator_ix=agent_ix,
            aea_key=self.service_spec.private_keys[agent_ix],
            number_of_validators=number_of_agents,
            host_names=host_names,
            tendermint_image_name=TENDERMINT_IMAGE_NAME,
            tendermint_image_version=image_versions["tendermint"],
            open_aea_image_name=OPEN_AEA_IMAGE_NAME,
            open_aea_image_version=image_versions["agent"],
        )
        agent_deployment_yaml = yaml.load_all(agent_deployment, Loader=yaml.FullLoader)  # type: ignore
        resources = []
        for resource in agent_deployment_yaml:
            if resource.get("kind") == "Deployment":
                for container in resource["spec"]["template"]["spec"]["containers"]:
                    if container["name"] == "aea":
                        container["env"] += [
                            {"name": k, "value": f"{v}"} for k, v in agent_vars.items()
                        ]
            resources.append(resource)

        res = "\n---\n".join([yaml.safe_dump(i) for i in resources])
        return res

    def generate_config_tendermint(
        self, image_version: str = TENDERMINT_IMAGE_VERSION
    ) -> "KubernetesGenerator":
        """Build configuration job."""

        if self.tendermint_job_config is not None:
            return self

        host_names = ", ".join(
            [
                f'"--hostname=abci{i}"'
                for i in range(self.service_spec.service.number_of_agents)
            ]
        )

        self.tendermint_job_config = CLUSTER_CONFIGURATION_TEMPLATE.format(
            valory_app=self.service_spec.service.agent.name,
            number_of_validators=self.service_spec.service.number_of_agents,
            host_names=host_names,
            tendermint_image_name=TENDERMINT_IMAGE_NAME,
            tendermint_image_version=image_version,
        )

        return self

    def _apply_cluster_specific_tendermint_params(  # pylint: disable= no-self-use
        self, agent_params: List
    ) -> List:
        """Override the tendermint params to point at the localhost."""
        for agent in agent_params:
            agent.update(TENDERMINT_CONFIGURATION_OVERRIDES[self.deployment_type])
        return agent_params

    def generate(  # pylint: disable=unused-argument
        self,
        image_versions: Dict[str, str],
        dev_mode: bool = False,
    ) -> "KubernetesGenerator":
        """Generate the deployment."""

        if dev_mode:
            self.resources.append(
                HARDHAT_TEMPLATE % (HARDHAT_IMAGE_NAME, image_versions["hardhat"])
            )

        agent_vars = self.service_spec.generate_agents()  # type:ignore
        agent_vars = self._apply_cluster_specific_tendermint_params(agent_vars)
        agent_vars = self.get_deployment_network_configuration(agent_vars)
        self.image_name = self.service_spec.service.agent.name

        agents = "\n---\n".join(
            [
                self.build_agent_deployment(
                    self.image_name,
                    i,
                    self.service_spec.service.number_of_agents,
                    agent_vars[i],
                    image_versions,
                )
                for i in range(self.service_spec.service.number_of_agents)
            ]
        )
        self.resources.append(agents)
        self.output = "\n---\n".join(self.resources)
        return self

    def write_config(
        self,
    ) -> "KubernetesGenerator":
        """Write output to build dir"""

        output = "---\n".join([self.output, cast(str, self.tendermint_job_config)])
        if not self.build_dir.is_dir():
            self.build_dir.mkdir()
        with open(
            self.build_dir / self.output_name, "w", encoding=DEFAULT_ENCODING
        ) as f:
            f.write(output)

        return self

    def populate_private_keys(self) -> "BaseDeploymentGenerator":
        """Populates private keys into a config map for the kubernetes deployment."""
        path = self.build_dir / "agent_keys"
        for x in range(self.service_spec.service.number_of_agents):
            key = self.service_spec.private_keys[x]
            secret = AGENT_SECRET_TEMPLATE.format(private_key=key, validator_ix=x)
            with open(
                path / KUBERNETES_AGENT_KEY_NAME.format(agent_n=x), "w", encoding="utf8"
            ) as f:
                f.write(secret)
        return self



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: contract_api.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x12\x63ontract_api.proto\x12\x1e\x61\x65\x61.valory.contract_api.v1_0_0"\x84\x11\n\x12\x43ontractApiMessage\x12V\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x45.aea.valory.contract_api.v1_0_0.ContractApiMessage.Error_PerformativeH\x00\x12x\n\x16get_deploy_transaction\x18\x06 \x01(\x0b\x32V.aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_Deploy_Transaction_PerformativeH\x00\x12j\n\x0fget_raw_message\x18\x07 \x01(\x0b\x32O.aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_Raw_Message_PerformativeH\x00\x12r\n\x13get_raw_transaction\x18\x08 \x01(\x0b\x32S.aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_Raw_Transaction_PerformativeH\x00\x12^\n\tget_state\x18\t \x01(\x0b\x32I.aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_State_PerformativeH\x00\x12\x62\n\x0braw_message\x18\n \x01(\x0b\x32K.aea.valory.contract_api.v1_0_0.ContractApiMessage.Raw_Message_PerformativeH\x00\x12j\n\x0fraw_transaction\x18\x0b \x01(\x0b\x32O.aea.valory.contract_api.v1_0_0.ContractApiMessage.Raw_Transaction_PerformativeH\x00\x12V\n\x05state\x18\x0c \x01(\x0b\x32\x45.aea.valory.contract_api.v1_0_0.ContractApiMessage.State_PerformativeH\x00\x1a\x18\n\x06Kwargs\x12\x0e\n\x06kwargs\x18\x01 \x01(\x0c\x1a!\n\nRawMessage\x12\x13\n\x0braw_message\x18\x01 \x01(\x0c\x1a)\n\x0eRawTransaction\x12\x17\n\x0fraw_transaction\x18\x01 \x01(\x0c\x1a\x16\n\x05State\x12\r\n\x05state\x18\x01 \x01(\x0c\x1a\xaa\x01\n#Get_Deploy_Transaction_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x10\n\x08\x63\x61llable\x18\x03 \x01(\t\x12I\n\x06kwargs\x18\x04 \x01(\x0b\x32\x39.aea.valory.contract_api.v1_0_0.ContractApiMessage.Kwargs\x1a\xc1\x01\n Get_Raw_Transaction_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x18\n\x10\x63ontract_address\x18\x03 \x01(\t\x12\x10\n\x08\x63\x61llable\x18\x04 \x01(\t\x12I\n\x06kwargs\x18\x05 \x01(\x0b\x32\x39.aea.valory.contract_api.v1_0_0.ContractApiMessage.Kwargs\x1a\xbd\x01\n\x1cGet_Raw_Message_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x18\n\x10\x63ontract_address\x18\x03 \x01(\t\x12\x10\n\x08\x63\x61llable\x18\x04 \x01(\t\x12I\n\x06kwargs\x18\x05 \x01(\x0b\x32\x39.aea.valory.contract_api.v1_0_0.ContractApiMessage.Kwargs\x1a\xb7\x01\n\x16Get_State_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x18\n\x10\x63ontract_address\x18\x03 \x01(\t\x12\x10\n\x08\x63\x61llable\x18\x04 \x01(\t\x12I\n\x06kwargs\x18\x05 \x01(\x0b\x32\x39.aea.valory.contract_api.v1_0_0.ContractApiMessage.Kwargs\x1a]\n\x12State_Performative\x12G\n\x05state\x18\x01 \x01(\x0b\x32\x38.aea.valory.contract_api.v1_0_0.ContractApiMessage.State\x1az\n\x1cRaw_Transaction_Performative\x12Z\n\x0fraw_transaction\x18\x01 \x01(\x0b\x32\x41.aea.valory.contract_api.v1_0_0.ContractApiMessage.RawTransaction\x1an\n\x18Raw_Message_Performative\x12R\n\x0braw_message\x18\x01 \x01(\x0b\x32=.aea.valory.contract_api.v1_0_0.ContractApiMessage.RawMessage\x1an\n\x12\x45rror_Performative\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x13\n\x0b\x63ode_is_set\x18\x02 \x01(\x08\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x16\n\x0emessage_is_set\x18\x04 \x01(\x08\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x42\x0e\n\x0cperformativeb\x06proto3'
)


_CONTRACTAPIMESSAGE = DESCRIPTOR.message_types_by_name["ContractApiMessage"]
_CONTRACTAPIMESSAGE_KWARGS = _CONTRACTAPIMESSAGE.nested_types_by_name["Kwargs"]
_CONTRACTAPIMESSAGE_RAWMESSAGE = _CONTRACTAPIMESSAGE.nested_types_by_name["RawMessage"]
_CONTRACTAPIMESSAGE_RAWTRANSACTION = _CONTRACTAPIMESSAGE.nested_types_by_name[
    "RawTransaction"
]
_CONTRACTAPIMESSAGE_STATE = _CONTRACTAPIMESSAGE.nested_types_by_name["State"]
_CONTRACTAPIMESSAGE_GET_DEPLOY_TRANSACTION_PERFORMATIVE = (
    _CONTRACTAPIMESSAGE.nested_types_by_name["Get_Deploy_Transaction_Performative"]
)
_CONTRACTAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE = (
    _CONTRACTAPIMESSAGE.nested_types_by_name["Get_Raw_Transaction_Performative"]
)
_CONTRACTAPIMESSAGE_GET_RAW_MESSAGE_PERFORMATIVE = (
    _CONTRACTAPIMESSAGE.nested_types_by_name["Get_Raw_Message_Performative"]
)
_CONTRACTAPIMESSAGE_GET_STATE_PERFORMATIVE = _CONTRACTAPIMESSAGE.nested_types_by_name[
    "Get_State_Performative"
]
_CONTRACTAPIMESSAGE_STATE_PERFORMATIVE = _CONTRACTAPIMESSAGE.nested_types_by_name[
    "State_Performative"
]
_CONTRACTAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE = (
    _CONTRACTAPIMESSAGE.nested_types_by_name["Raw_Transaction_Performative"]
)
_CONTRACTAPIMESSAGE_RAW_MESSAGE_PERFORMATIVE = _CONTRACTAPIMESSAGE.nested_types_by_name[
    "Raw_Message_Performative"
]
_CONTRACTAPIMESSAGE_ERROR_PERFORMATIVE = _CONTRACTAPIMESSAGE.nested_types_by_name[
    "Error_Performative"
]
ContractApiMessage = _reflection.GeneratedProtocolMessageType(
    "ContractApiMessage",
    (_message.Message,),
    {
        "Kwargs": _reflection.GeneratedProtocolMessageType(
            "Kwargs",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_KWARGS,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Kwargs)
            },
        ),
        "RawMessage": _reflection.GeneratedProtocolMessageType(
            "RawMessage",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_RAWMESSAGE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.RawMessage)
            },
        ),
        "RawTransaction": _reflection.GeneratedProtocolMessageType(
            "RawTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_RAWTRANSACTION,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.RawTransaction)
            },
        ),
        "State": _reflection.GeneratedProtocolMessageType(
            "State",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_STATE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.State)
            },
        ),
        "Get_Deploy_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Deploy_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_GET_DEPLOY_TRANSACTION_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_Deploy_Transaction_Performative)
            },
        ),
        "Get_Raw_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Raw_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_Raw_Transaction_Performative)
            },
        ),
        "Get_Raw_Message_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Raw_Message_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_GET_RAW_MESSAGE_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_Raw_Message_Performative)
            },
        ),
        "Get_State_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_State_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_GET_STATE_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Get_State_Performative)
            },
        ),
        "State_Performative": _reflection.GeneratedProtocolMessageType(
            "State_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_STATE_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.State_Performative)
            },
        ),
        "Raw_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Raw_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Raw_Transaction_Performative)
            },
        ),
        "Raw_Message_Performative": _reflection.GeneratedProtocolMessageType(
            "Raw_Message_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_RAW_MESSAGE_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Raw_Message_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONTRACTAPIMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "contract_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _CONTRACTAPIMESSAGE,
        "__module__": "contract_api_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.contract_api.v1_0_0.ContractApiMessage)
    },
)
_sym_db.RegisterMessage(ContractApiMessage)
_sym_db.RegisterMessage(ContractApiMessage.Kwargs)
_sym_db.RegisterMessage(ContractApiMessage.RawMessage)
_sym_db.RegisterMessage(ContractApiMessage.RawTransaction)
_sym_db.RegisterMessage(ContractApiMessage.State)
_sym_db.RegisterMessage(ContractApiMessage.Get_Deploy_Transaction_Performative)
_sym_db.RegisterMessage(ContractApiMessage.Get_Raw_Transaction_Performative)
_sym_db.RegisterMessage(ContractApiMessage.Get_Raw_Message_Performative)
_sym_db.RegisterMessage(ContractApiMessage.Get_State_Performative)
_sym_db.RegisterMessage(ContractApiMessage.State_Performative)
_sym_db.RegisterMessage(ContractApiMessage.Raw_Transaction_Performative)
_sym_db.RegisterMessage(ContractApiMessage.Raw_Message_Performative)
_sym_db.RegisterMessage(ContractApiMessage.Error_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CONTRACTAPIMESSAGE._serialized_start = 55
    _CONTRACTAPIMESSAGE._serialized_end = 2235
    _CONTRACTAPIMESSAGE_KWARGS._serialized_start = 903
    _CONTRACTAPIMESSAGE_KWARGS._serialized_end = 927
    _CONTRACTAPIMESSAGE_RAWMESSAGE._serialized_start = 929
    _CONTRACTAPIMESSAGE_RAWMESSAGE._serialized_end = 962
    _CONTRACTAPIMESSAGE_RAWTRANSACTION._serialized_start = 964
    _CONTRACTAPIMESSAGE_RAWTRANSACTION._serialized_end = 1005
    _CONTRACTAPIMESSAGE_STATE._serialized_start = 1007
    _CONTRACTAPIMESSAGE_STATE._serialized_end = 1029
    _CONTRACTAPIMESSAGE_GET_DEPLOY_TRANSACTION_PERFORMATIVE._serialized_start = 1032
    _CONTRACTAPIMESSAGE_GET_DEPLOY_TRANSACTION_PERFORMATIVE._serialized_end = 1202
    _CONTRACTAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE._serialized_start = 1205
    _CONTRACTAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE._serialized_end = 1398
    _CONTRACTAPIMESSAGE_GET_RAW_MESSAGE_PERFORMATIVE._serialized_start = 1401
    _CONTRACTAPIMESSAGE_GET_RAW_MESSAGE_PERFORMATIVE._serialized_end = 1590
    _CONTRACTAPIMESSAGE_GET_STATE_PERFORMATIVE._serialized_start = 1593
    _CONTRACTAPIMESSAGE_GET_STATE_PERFORMATIVE._serialized_end = 1776
    _CONTRACTAPIMESSAGE_STATE_PERFORMATIVE._serialized_start = 1778
    _CONTRACTAPIMESSAGE_STATE_PERFORMATIVE._serialized_end = 1871
    _CONTRACTAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE._serialized_start = 1873
    _CONTRACTAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE._serialized_end = 1995
    _CONTRACTAPIMESSAGE_RAW_MESSAGE_PERFORMATIVE._serialized_start = 1997
    _CONTRACTAPIMESSAGE_RAW_MESSAGE_PERFORMATIVE._serialized_end = 2107
    _CONTRACTAPIMESSAGE_ERROR_PERFORMATIVE._serialized_start = 2109
    _CONTRACTAPIMESSAGE_ERROR_PERFORMATIVE._serialized_end = 2219
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains utilities for the 'counter_client' skill."""
# isort:skip_file # noqa
import base64
import datetime
import json
import struct

from packages.valory.protocols.http import (  # type: ignore # pylint: disable=no-name-in-module,import-error
    HttpMessage,
)


def decode_value(message: HttpMessage) -> int:
    """Decode the counter value."""
    content_bytes = message.body
    content = json.loads(content_bytes)
    counter_value_base64 = content["result"]["response"]["value"].encode("ascii")
    counter_value_bytes = base64.b64decode(counter_value_base64)
    counter_value, *_ = struct.unpack(">I", counter_value_bytes)
    return counter_value


def curdatetime() -> str:
    """Return current datetime in isoformat.

    This is a local method that does not depend on the global clock,
    so the usage of datetime.now() is not a problem here.

    :return: the current time in isoformat.
    """
    datetime.datetime.utcnow()
    return datetime.datetime.now().isoformat()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base configurations."""

from collections import OrderedDict
from copy import copy
from pathlib import Path
from typing import Any, Dict, FrozenSet, List, Optional, Sequence, Tuple, cast

from aea import AEA_DIR
from aea.configurations import validation
from aea.configurations.base import (
    ComponentConfiguration,
    ComponentId,
    ConnectionConfig,
    ContractConfig,
    PackageConfiguration,
    ProtocolConfig,
    SkillConfig,
)
from aea.configurations.constants import CONNECTION, CONTRACT, PROTOCOL, SKILL
from aea.configurations.data_types import PackageType, PublicId
from aea.helpers.base import SimpleIdOrStr, cd

from autonomy.configurations.constants import DEFAULT_SERVICE_FILE, SCHEMAS_DIR
from autonomy.configurations.validation import ConfigValidator


COMPONENT_CONFIGS: Dict = {
    component.package_type.value: component  # type: ignore
    for component in [
        ContractConfig,
        SkillConfig,
        ProtocolConfig,
        ConnectionConfig,
    ]
}


def recurse(obj: Dict[str, Any]) -> Dict[str, Any]:
    """Recursively explore a json object until no dictionaries remain."""
    if not any([isinstance(i, dict) for i in obj.values()]):
        return obj

    new_obj = {}
    for k, v in obj.items():
        if isinstance(v, dict):
            for k2, v2 in v.items():
                new_obj["_".join([str(k), str(k2)])] = v2
        else:
            new_obj[k] = v
    return recurse(new_obj)


def _parse_nested_override(env_var_name: str, nested_override_value: Dict) -> Dict:
    """Used for handling dictionary object 1 level below nesting."""
    overrides = recurse(nested_override_value)
    return {f"{env_var_name}_{k}".upper(): v for k, v in overrides.items()}


class Service(PackageConfiguration):  # pylint: disable=too-many-instance-attributes
    """Service package configuration."""

    default_configuration_filename = DEFAULT_SERVICE_FILE
    package_type = PackageType.SERVICE
    schema = str(SCHEMAS_DIR.absolute() / "service_schema.json")

    FIELDS_ALLOWED_TO_UPDATE: FrozenSet[str] = frozenset()

    __slots__ = (
        "_name",
        "_author",
        "version",
        "license",
        "fingerprint",
        "fingerprint_ignore_patterns",
        "build_entrypoint",
        "agent",
        "network",
        "number_of_agents",
        "_aea_version",
        "_aea_version_specifiers",
        "_directory",
        "_overrides",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: SimpleIdOrStr,
        author: SimpleIdOrStr,
        agent: PublicId,
        version: str = "",
        license_: str = "",
        aea_version: str = "",
        fingerprint: Optional[Dict[str, str]] = None,
        fingerprint_ignore_patterns: Optional[Sequence[str]] = None,
        description: str = "",
        number_of_agents: int = 4,
        network: Optional[str] = None,
        build_entrypoint: Optional[str] = None,
        overrides: Optional[List] = None,
    ) -> None:
        """Initialise object."""

        super().__init__(
            name=name,
            author=author,
            version=version,
            license_=license_,
            aea_version=aea_version,
            build_entrypoint=build_entrypoint,
            fingerprint=fingerprint,
            fingerprint_ignore_patterns=fingerprint_ignore_patterns,
        )

        self.agent = PublicId.from_str(str(agent))
        self.description = description
        self.number_of_agents = number_of_agents
        self.network = network

        self._overrides = [] if overrides is None else overrides

    @property
    def overrides(
        self,
    ) -> List:
        """Returns component overrides."""

        return self._overrides

    @overrides.setter
    def overrides(self, obj: List) -> None:
        """Set overrides."""

        self.check_overrides_valid(obj)
        self.check_overrides_match_spec(obj)
        self._overrides = obj

    @property
    def json(
        self,
    ) -> Dict:
        """Returns an ordered Dict for service config."""

        config = OrderedDict(
            {
                "name": self.name,
                "author": self.author,
                "agent": str(self.agent),
                "version": self.version,
                "license": self.license,
                "aea_version": self.aea_version,
                "description": self.description,
                "number_of_agents": self.number_of_agents,
                "network": self.network,
                "overrides": self.overrides,
                "fingerprint": self.fingerprint,
                "fingerprint_ignore_patterns": self.fingerprint_ignore_patterns,
            }
        )

        return config

    @classmethod
    def _create_or_update_from_json(cls, obj: Dict, instance: Any = None) -> "Service":
        """Create or update from json data."""

        obj = {**(instance.json if instance else {}), **copy(obj)}
        params = dict(
            name=cast(str, obj.get("name")),
            author=cast(str, obj.get("author")),
            agent=cast(str, obj.get("agent")),
            version=cast(str, obj.get("version")),
            license_=cast(str, obj.get("license")),
            aea_version=cast(str, obj.get("aea_version")),
            description=cast(str, obj.get("description")),
            number_of_agents=cast(int, obj.get("number_of_agents")),
            network=cast(str, obj.get("network")),
            overrides=cast(List, obj.get("overrides", [])),
            fingerprint=cast(Dict[str, str], obj.get("fingerprint", [])),
            fingerprint_ignore_patterns=cast(
                Sequence[str], obj.get("fingerprint_ignore_patterns", [])
            ),
        )

        return cls(**params)  # type: ignore

    @classmethod
    def validate_config_data(
        cls, json_data: Dict, env_vars_friendly: bool = False
    ) -> None:
        """Validate config data."""
        ConfigValidator(cls.schema, env_vars_friendly=env_vars_friendly).validate(
            json_data
        )

    def check_overrides_match_spec(self, overrides: List) -> bool:
        """Check that overrides are valid.

        - number of overrides is 1
        - number of overrides == number of agents in spec
        - number of overrides is 0

        :param overrides: List of overrides
        :return: True if overrides are valid
        """
        valid = []
        remaining = copy(overrides)

        for component in [
            CONNECTION,
            CONTRACT,
            PROTOCOL,
            SKILL,
        ]:

            component_overrides = [f for f in overrides if f["type"] == component]
            remaining = [f for f in remaining if f not in component_overrides]
            if any(
                [
                    self.number_of_agents == len(component_overrides),
                    len(component_overrides) == 0,
                    len(component_overrides) == 1,
                ]
            ):
                valid.append(True)

        if len(remaining) > 0:
            raise ValueError(f"Override type is misspelled.\n {remaining}")

        if sum(valid) == 4:
            return True

        raise ValueError("Incorrect number of overrides for count of agents.")

    def check_overrides_valid(
        self, overrides: List, env_vars_friendly: bool = False
    ) -> Dict[ComponentId, Dict[Any, Any]]:
        """Uses the aea helper libraries to check individual overrides."""

        component_configurations: Dict[ComponentId, Dict] = {}
        # load the other components.

        for idx, component_configuration_json in enumerate(overrides):
            component_id, _ = self.process_component_section(
                idx, component_configuration_json
            )
            if component_id in component_configurations:
                raise ValueError(
                    f"Configuration of component {component_id} occurs more than once."
                )
            component_configurations[component_id] = component_configuration_json

        return component_configurations

    def process_component_section(
        self,
        component_index: int,
        component_configuration_json: Dict,
    ) -> Tuple[ComponentId, Dict]:
        """
        Process a component configuration in an agent configuration file.

        It breaks down in:
        - extract the component id
        - validate the component configuration
        - check that there are only configurable fields

        :param component_index: the index of the component in the file.
        :param component_configuration_json: the JSON object.
        :return: the processed component configuration.
        """
        configuration = copy(component_configuration_json)
        component_id = ConfigValidator.split_component_id_and_config(
            component_index, configuration
        )

        path = Path(AEA_DIR) / "configurations" / "schemas"
        config_class = COMPONENT_CONFIGS[component_id.package_type.value]

        with cd(path):  # required to handle protected variable _SCHEMEAS_DIR
            cv = validation.ConfigValidator("definitions.json")
            try:
                cv.validate_component_configuration(component_id, configuration)
                overrides = self.try_to_process_singular_override(
                    component_id, config_class, configuration
                )
            except ValueError as e:
                print(
                    f"Failed to parse as a singular input with {e}\nAttempting with nested fields."
                )

                overrides = self.try_to_process_nested_fields(
                    component_id,
                    component_index,
                    config_class,
                    configuration,
                )
        return component_id, overrides

    @staticmethod
    def try_to_process_singular_override(
        component_id: ComponentId,
        config_class: ComponentConfiguration,
        component_configuration_json: Dict,
    ) -> Dict:
        """Try to process component with a singular component overrides."""
        overrides = {}
        for field in config_class.FIELDS_ALLOWED_TO_UPDATE:
            env_var_base = "_".join(
                [component_id.package_type.value, component_id.name, field]
            )

            field_override = component_configuration_json.get(field, {})
            if field_override == {}:
                continue
            for nested_override, nested_value in field_override.items():
                for (
                    nested_override_key,
                    nested_override_value,
                ) in nested_value.items():
                    env_var_name = "_".join(
                        [env_var_base, nested_override, nested_override_key]
                    )
                    overrides.update(
                        _parse_nested_override(env_var_name, nested_override_value)
                    )
        return overrides

    def try_to_process_nested_fields(  # pylint: disable=too-many-locals
        self,
        component_id: ComponentId,
        component_index: int,
        config_class: ComponentConfiguration,
        component_configuration_json: Dict,
    ) -> Dict:
        """Try to process component with nested overrides."""
        overrides = {}
        for field in config_class.FIELDS_WITH_NESTED_FIELDS:  # type: ignore
            field_override = component_configuration_json.get(field, {})
            if field_override == {}:
                continue
            if not all(isinstance(item, int) for item in field_override.keys()):
                raise ValueError(
                    "All keys of list like override should be of type int."
                )
            nums = set(field_override.keys())

            if len(nums) != self.number_of_agents:
                raise ValueError("Not enough items in override")

            if nums != set(range(0, self.number_of_agents)):
                raise ValueError("Overrides incorrectly indexed")

            n_fields = len(field_override)
            for override in field_override[component_index % n_fields]:
                for nested_override, nested_value in override.items():
                    for (
                        nested_override_key,
                        nested_override_value,
                    ) in nested_value.items():
                        if (
                            nested_override_key
                            not in config_class.NESTED_FIELDS_ALLOWED_TO_UPDATE  # type: ignore
                        ):
                            raise ValueError("Trying to override non-nested field.")

                        env_var_name = "_".join(
                            [
                                component_id.package_type.value,
                                component_id.name,
                                field,
                                nested_override,
                                nested_override_key,
                            ]
                        )
                        overrides.update(
                            _parse_nested_override(env_var_name, nested_override_value)
                        )

        return overrides



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the models.py module of the skill."""

from packages.valory.skills.liquidity_rebalancing_abci.models import Params, SharedState


class DummyContext:
    """Dummy Context class for shared state."""

    class params:
        """Dummy param variable."""

        round_timeout_seconds: float = 1.0


class TestSharedState:
    """Test SharedState(Model) class."""

    def test_initialization(
        self,
    ) -> None:
        """Test initialization."""
        SharedState(name="", skill_context=DummyContext())


class TestParams:
    """Test params."""

    def test_initialization(
        self,
    ) -> None:
        """Test initialization."""
        Params(
            name="",
            skill_context=DummyContext(),
            consensus={"max_participants": 1},
            max_healthcheck=10,
            round_timeout_seconds=1,
            sleep_time=1,
            retry_timeout=1,
            retry_attempts=1,
            validate_timeout=1,
            finalize_timeout=1,
            history_check_timeout=1,
            drand_public_key="868f005eb8e6e4ca0a47c8a77ceaa5309a47978a7c71bc5cce96366b5d7a569937c529eeda66c7293784a9402801af31",
            observation_interval=10,
            setup={},
            reset_tendermint_after=2,
            tendermint_com_url="http://localhost:8080",
            tendermint_url="http://localhost:26657",
            tendermint_max_retries=5,
            tendermint_check_sleep_delay=3,
            service_id="liquidity_rebalancing",
            keeper_timeout=1.0,
            keeper_allowed_retries=3,
            rebalancing={
                "chain": "Ethereum",
                "token_base_address": "0xDc64a140Aa3E981100a9becA4E685f962f0cF6C9",
                "token_base_ticker": "WETH",
                "token_a_address": "0x0DCd1Bf9A1b36cE34237eEaFef220932846BCD82",
                "token_a_ticker": "TKA",
                "token_b_address": "0x9A676e781A523b5d0C0e43731313A708CB607508",
                "token_b_ticker": "TKB",
                "lp_token_address": "0x50CD56fb094F8f06063066a619D898475dD3EedE",
                "default_minter": "0x0000000000000000000000000000000000000000",
                "ab_pool_address": "0x86A6C37D3E868580a65C723AAd7E0a945E170416",
                "max_allowance": 1.1579209e77,  # 2**256 - 1
                "deadline": 300,  # 5 min into the future"
                "sleep_seconds": 60,
            },
            cleanup_history_depth=0,
        )
